#include 'directry.ch'
#include 'common.ch'

procedure local( filespec )

local aFileNames
local nNumFiles
local cThisFile
local cThisLine
local cThisCommand
local x
local y
local z
local outhandle
local nPadSpace
local aLocals
local nPointer
local cLocal
local lWorkDone
local nCmt1
local nCmt2
local cComment
local lStatFunc
local cOutFile
local cTempVar
local aPreCombined
local nPreCombined

default filespec to '*.prg'

// Grab the files to work on

aFileNames := directory( filespec )

// Find out how many files to work with

nNumFiles := len( aFileNames )

// process all files

? 'Declaration Buster,  Version 1.1'
?
? 'Working set: ' + filespec
?

for x := 1 to nNumFiles

   // grab the file name from the working array

   cThisFile := afileNames[ x, F_NAME ]

   ? 'Processing => ' + pad( cThisFile, 12 )

   cOutFile := rootname( cThisFile ) + '.FIX'

   // does the output file exist? If so, skip it
   if file( cOutFile )
      ?? '   Skip. Output file name already exists.'
      loop
   endif

   lWorkDone := .f.

   // open the file for buffered reading
   if bopen( cThisFile )

      // create a file by the same name, with a .FIX extension

      outhandle := fcreate( cOutFile, 0 )

      // work on this file until all lines are read

      do while !beof()

         // get the next line from the buffered read

         cThisLine := bReadLine()

         // check to see if the line is a declaration line

         if left( upper( alltrim( cThisLine ) ), 4 ) $ ;
                                         'LOCA|PRIV|MEMV|STAT|PUBL'

            // this determines how many spaces were padded
            // on the left of the declaration, and strips them off

            nPadSpace := 0
            do while left( cThisLine, 1 ) == ' '
               nPadSpace ++
               cThisLine := substr( cThisLine, 2 )
            enddo

            // this checks to see if this line has a continuation -> ;
            // and if so it grabs the rest of the pieces and
            // creates a single long line out of them.

            // it also gathers an array of all the lines before they
            // get combined, so if there is nothing to do, we can
            // put them back the way they started.

            cThisline := alltrim( cThisline )
            aPreCombined := { cThisLine }
            nPreCombined := 1
            do while right( cThisLine, 1 ) == ';'
               cThisLine := left( cThisLine, len( cThisLine ) - 1 )
               cTempVar := bReadLine()
               aadd( aPreCombined, cTempVar )
               nPreCombined ++
               cThisLine += alltrim( cTempVar )
            enddo

            // now, if there are end of line comments, we need to
            // chop them off of the working lines

            nCmt1 := at( '//', cThisline )
            nCmt2 := at( '&&', cThisLine )

            if !empty( nCmt1 ) .and. !empty( nCmt2 )
               nPointer := min( nCmt1, nCmt2 )
            else
               nPointer := max( nCmt1, nCmt2 )
            endif

            if nPointer == 0
               cComment := ''
            else
               cComment := ' ' + substr( cThisLine, nPointer )
               cThisLine := trim( left( cThisLine, nPointer - 1 ) )
            endif

            // check for STATIC FUNCTION's, yet another overloaded term!
            lStatFunc := .f.
            // find out where second word starts
            nPointer := at( ' ', cThisLine )
            if upper( left( cThisline, 4 ) ) == 'STAT'
               if upper( left( substr( cThisline, nPointer + 1 ), 4 ) ) == 'FUNC'
                  lStatFunc := .t.
               endif
            endif

            // check to see if we can skip this declaration
            // because it only has one part

            if at( ',', cThisLine ) == 0 .or. ;
               nextcomma( cThisLine ) == 0 .or. ;
               lStatFunc

               // No need to break this one up, so just write it
               // from data gathered before it was assembled from
               // multiple lines.

               for z := 1 to nPreCombined
                  fwrite( outhandle, aPreCombined[ z ] + chr( 13 ) + chr( 10 ) )
               next

            else

               // once we get to here, we have found a line which
               // needs to be broken apart.

               // first, we grab the declaration command off of the
               // left end of the line.

               cThisCommand := left( cThisLine, at( ' ', cThisLine ) - 1 )
               cThisLine := substr( cThisLine, len( cThisCommand ) + 1 )

               // now, we go throught the line and seperate each
               // declaration into the aLocals array

               aLocals := {}
               do while .t.

                  // this gets the next divider comma

                  nPointer := nextcomma( cThisLine )

                  // if nPoiner is still greater than 0, there are still
                  // more parts to seperate

                  if nPointer > 0

                     // grab off this piece and put it on the array

                     aadd( aLocals, left( cThisLine, nPointer - 1 ) )

                     // then trim it off of the working line

                     cThisLine := substr( cThisLine, nPointer + 1 )

                  else

                     // this is the last one!

                     aadd( aLocals, cThisLine )
                     exit

                  endif

               enddo

               // now, write the output to the file in seperate pieces

               for y := 1 to len( aLocals )
                  fwrite( outhandle, space( nPadSpace ) + ;
                                     cThisCommand + ;
                                     ' ' + ;
                                     alltrim( aLocals[ y ] ) + ;
                                     cComment + ;
                                     chr( 13 ) + chr( 10 ) )
               next

               lWorkDone := .t.

            endif

         else

            // this line is not a declaration line, write it with no change

            fwrite( outhandle, cThisLine + chr( 13 ) + chr( 10 ) )

         endif

      enddo

      // we are done with this file, so close the input and output

      bclose()
      fclose( outhandle )

   endif

   if lWorkDone
      ?? '   Done. Output is ' + rootname( cThisFile ) + '.FIX'
   else
      ?? '   Done. No Fix, no output'
      ferase( rootname( cThisFile ) + '.FIX' )
   endif

next

return

*******************************************************************

/*

Realistically, this is the heartbeat of this entire process. Without this
function, it is impossible to determine where a multiple declaration begins
and ends, since Clipper language allows overloading of symbols in the
language. In this case, the comma is overloaded. It means more than
one thing.

Rules:

Anything inside an inline comment is bypassed

Any comma inside a function should be bypassed

Comma, Parenthesis, Braces and Brackets should be bypassed if they are
inside a string declaration.

Single quotes are bypassed if they are inside
a double quote string declaration

Double quotes are bypassed if they are inside
a single quote string declaration

Note:

This routine does not consider [ ] characters to be string declaration.
Although they can legally be used that way, I have not found it to be
common. It also difficult to determine when they are being used in this
manner.

*/

function nextcomma( cThisLine )

local nPointer := 0
local nParen := 0
local nBrack := 0
local nBrace := 0
local nSQuote := 0
local nDQuote := 0
local inlinecmt := 0
local x
local nNumChars := len( cThisLine )
local cThisChar
local cTwoChars

for x := 1 to nNumChars

   cTwoChars := substr( cThisLine, x, 2 )

   do case
   case cTwoChars == '/*'
      inlinecmt ++
   case cTwoChars == '*/'
      inlinecmt --
   endcase

   if inlinecmt == 0

      cThisChar := substr( cThisLine, x, 1 )

      if nSQuote == 0 .and. nDQuote == 0
         do case
         case cThisChar == '('
            nParen ++
         case cThisChar == ')'
            nParen --
         case cThisChar == '{'
            nBrace ++
         case cThisChar == '}'
            nBrace --
         case cThisChar == '['
            nBrack ++
         case cThisChar == ']'
            nBrack --
         endcase
      endif

      do case
      case cThisChar == '"'
         if nSQuote == 0
            if nDQuote == 1
               nDQuote := 0
            else
               nDQuote := 1
            endif
         endif
      case cThisChar == "'"
         if nDQuote == 0
            if nSQuote == 1
               nSQuote := 0
            else
               nSQuote := 1
            endif
         endif
      endcase

      // if all of this is true, then we just found a comma worth dividing on!

      if cThisChar == ',' .and. ;
            nParen ==  0  .and. ;
            nBrack ==  0  .and. ;
            nBrace ==  0  .and. ;
           nSQuote ==  0  .and. ;
           nDQuote ==  0

         nPointer := x
         exit

      endif

   endif

next

return nPointer

*******************************************************************

/* This function removes the extension from a filename */

function ROOTNAME( FILE_NAME )

local DOTAT := rat( '.', FILE_NAME )  // make sure we get the last dot!

if DOTAT > 0
  FILE_NAME := left( FILE_NAME, DOTAT - 1 )
endif

return alltrim( FILE_NAME )

